from ..defs import Polarity

from . import conjugators
from .defs import AdjectiveForm, AdjectiveClass

def conjugate(
    adj_dictform: str,
    adj_class: AdjectiveClass,
    form: AdjectiveForm,
    polarity: Polarity):
  """
  Returns the conjugated adjective.
  """

  CONJUGATORS = {
    AdjectiveForm.Plain: conjugators.plain,
    AdjectiveForm.PlainPast: conjugators.plainpast,
    AdjectiveForm.Polite: conjugators.polite,
    AdjectiveForm.PolitePast: conjugators.politepast,
    AdjectiveForm.Connective: conjugators.connective
  }
  return CONJUGATORS[form](adj_dictform, adj_class, polarity)


from ..defs import Polarity
from .defs import VerbClass, VerbForm
from . import conjugators

def conjugate(
    verb_dictform: str,
    verb_class: VerbClass,
    form: VerbForm,
    polarity: Polarity):
  """
  Calls the appropriate conjugator in jp_verb_conjugator.conjugators, and returns the conjugated verb.
  """
  CONJUGATORS = {
    VerbForm.Te: conjugators.teform,
    VerbForm.Plain: conjugators.plain,
    VerbForm.PlainPast: conjugators.plainpast,
    VerbForm.Polite: conjugators.polite,
    VerbForm.PolitePast: conjugators.politepast
  }
  return CONJUGATORS[form](verb_dictform, verb_class, polarity)


from ..defs import Polarity
from .defs import AdjectiveClass
from .conjugation_helpers import common_logic


def plain(adj_dictform, adj_class: AdjectiveClass, polarity: Polarity):
  return common_logic(adj_dictform, adj_class, polarity, 
    affirmative_i_suffix=adj_dictform[-1], # restore last mora
    negative_suffix="い")


def plainpast(adj_dictform, adj_class: AdjectiveClass, polarity: Polarity):
  return common_logic(adj_dictform, adj_class, polarity, 
    affirmative_i_suffix="かった",
    affirmative_na_suffix="だった",
    negative_suffix="かった")


def polite(adj_dictform, adj_class: AdjectiveClass, polarity: Polarity):
  return plain(adj_dictform, adj_class, polarity) + "です"


def politepast(adj_dictform, adj_class: AdjectiveClass, polarity: Polarity):
  return common_logic(adj_dictform, adj_class, polarity, 
    affirmative_i_suffix="かったです",
    affirmative_na_suffix="でした",
    negative_suffix="かったです")


def connective(adj_dictform, adj_class: AdjectiveClass, polarity: Polarity):
  return common_logic(adj_dictform, adj_class, polarity, 
    affirmative_i_suffix="くて",
    affirmative_na_suffix="で",
    negative_suffix="くて")



from . import conjugation_helpers as helpers
from .defs import VerbClass
from ..defs import Polarity


def plain(verb_dictform: str, verb_class: VerbClass, polarity: Polarity):
  if polarity == Polarity.Affirmative:
    return verb_dictform
  else:
    return helpers.nai_stem(verb_dictform, verb_class) + 'ない'


def plainpast(verb_dictform: str, verb_class: VerbClass, polarity: Polarity):
  if polarity == Polarity.Affirmative:
    te_form = helpers.te_form(verb_dictform, verb_class)
    last_mora = te_form[-1]
    
    if last_mora == 'て':
      return te_form[:-1] + 'た'
    elif last_mora == 'で':
      return te_form[:-1] + 'だ'
    else:
      assert False # impossible unless base_conjugators.te_form is broken
    
  else:
    return helpers.nai_stem(verb_dictform, verb_class) + 'なかった'


def polite(verb_dictform: str, verb_class: VerbClass, polarity: Polarity):
  stem = helpers.masu_stem(verb_dictform, verb_class)
  
  if polarity == Polarity.Affirmative:
    return stem + 'ます'
  else:
    return stem + 'ません'


def politepast(verb_dictform: str, verb_class: VerbClass, polarity: Polarity):
  stem = helpers.masu_stem(verb_dictform, verb_class)

  if polarity == Polarity.Affirmative:
    return stem + 'ました'
  else:
    return stem + 'ませんでした'
  

def teform(verb_dictform: str, verb_class: VerbClass, polarity: Polarity):
  if polarity == Polarity.Affirmative:
    return helpers.te_form(verb_dictform, verb_class)
  else:
    return helpers.nai_stem(verb_dictform, verb_class) + 'なくて'



from .defs import VerbClass

class WrongVerbClassException(Exception):
  def __init__(self, verb, verb_class):
    self.verb = verb
    self.verb_class = verb_class

    super().__init__(f'Wrong verb class "{verb_class}" for verb "{verb}".')

class UnsupportedVerbClassException(Exception):
  def __init__(self, verb_class):
    self.verb_class = verb_class
    super().__init__(f'Unsupported verb class "{verb_class}".')

def __ichidan_remove_ru(verb_dictform):
  # Remove る
  if verb_dictform[-1] != 'る':
    raise WrongVerbClassException(verb_dictform, VerbClass.Ichidan)
  
  return verb_dictform[:-1]

def __godan_map_last_mora(verb_dictform, mora_map: dict):
  # Transform the last mora with mora_map
  transformed = mora_map.get(verb_dictform[-1])
  if transformed is None:
    raise WrongVerbClassException(verb_dictform, VerbClass.Godan)
  
  return verb_dictform[:-1] + transformed


def masu_stem(verb_dictform, verb_class: VerbClass):
  if verb_class == VerbClass.Ichidan:
    # Remove る
    return __ichidan_remove_ru(verb_dictform)
  
  elif verb_class == VerbClass.Godan:
    # Transpose last mora to い-column
    MAP = {
      'う': 'い',
      'つ': 'ち',
      'る': 'り',
      'む': 'み',
      'ぶ': 'び',
      'ぬ': 'に',
      'く': 'き',
      'ぐ': 'ぎ',
      'す': 'し'
    }
    return __godan_map_last_mora(verb_dictform, MAP)
    
  else:
    raise UnsupportedVerbClassException(verb_class)


def nai_stem(verb_dictform, verb_class):
  if verb_class == VerbClass.Ichidan:
    # Remove る
    return __ichidan_remove_ru(verb_dictform)
  
  elif verb_class == VerbClass.Godan:
    # Transpose last mora to あ-column (except う goes to わ)
    MAP = {
      'う': 'わ',
      'つ': 'た',
      'る': 'ら',
      'む': 'ま',
      'ぶ': 'ば',
      'ぬ': 'な',
      'く': 'か',
      'ぐ': 'が',
      'す': 'さ'
    }
    return __godan_map_last_mora(verb_dictform, MAP)
    
  else:
    raise UnsupportedVerbClassException(verb_class)


def te_form(verb_dictform, verb_class):
  if verb_class == VerbClass.Ichidan:
    # Remove る
    return __ichidan_remove_ru(verb_dictform) + 'て'
  
  elif verb_class == VerbClass.Godan:
    # Apply godan Te-form rules
    MAP = {
      'う': 'って',
      'つ': 'って',
      'る': 'って',
      'む': 'んで',
      'ぶ': 'んで',
      'ぬ': 'んで',
      'く': 'いて',
      'ぐ': 'いで',
      'す': 'して'
    }
    return __godan_map_last_mora(verb_dictform, MAP)
    
  else:
    raise UnsupportedVerbClassException(verb_class)


from ..defs import Polarity
from .defs import AdjectiveClass

def invariant_with_suffix(adj_dictform, adj_class: AdjectiveClass, na_suffix = '', i_suffix = ''):
  if adj_class == AdjectiveClass.I:
    return adj_dictform[:-1] + i_suffix
  else:
    return adj_dictform + na_suffix

def negative_invariant(adj_dictform, adj_class: AdjectiveClass):
  return invariant_with_suffix(adj_dictform, adj_class, 
    i_suffix = 'くな',
    na_suffix = 'じゃな')

def common_logic(adj_dictform, adj_class: AdjectiveClass, polarity: Polarity,
    affirmative_i_suffix = '',
    affirmative_na_suffix = '',
    negative_suffix = ''):
  
  if polarity == Polarity.Affirmative:
    return invariant_with_suffix(adj_dictform, adj_class,
      i_suffix = affirmative_i_suffix,
      na_suffix = affirmative_na_suffix)
  else:
    return negative_invariant(adj_dictform, adj_class) + negative_suffix


import enum

class AdjectiveClass(enum.StrEnum):
  I = "i"
  Na = "na"

class AdjectiveForm(enum.StrEnum):
  Plain = "plain",
  PlainPast = "plainpast",
  Polite = "polite",
  PolitePast = "politepast",
  Connective = "connective"


import enum

class VerbClass(enum.StrEnum):
  Ichidan = "ichidan"
  Godan = "godan"

class VerbForm(enum.StrEnum):
  Plain = "plain",
  PlainPast = "plainpast",
  Polite = "polite",
  PolitePast = "politepast",
  Te = "te"